"""
MCP3xxx analog to digital converter via GPIOZero
"""

from ...exceptions import RuntimeConfigError
from ...types import ConfigType, SensorValueType
from . import GenericSensor

REQUIREMENTS = ("gpiozero",)
ALLOWED_TYPES = [
    "MCP3001",
    "MCP3002",
    "MCP3004",
    "MCP3008",
    "MCP3201",
    "MCP3202",
    "MCP3204",
    "MCP3208",
    "MCP3301",
    "MCP3302",
    "MCP3304",
    "MCP3308",
]
CONFIG_SCHEMA = {
    "spi_port": dict(type="integer", required=False, empty=False, default=0),
    "spi_device": dict(type="integer", required=False, empty=False, default=0),
    "type": dict(
        type="string",
        required=True,
        empty=False,
        allowed=ALLOWED_TYPES + [x.lower() for x in ALLOWED_TYPES],
    ),
    "channel": dict(type="integer", required=False, empty=False, default=0),
    "differential": dict(type="boolean", required=False, empty=False, default=False),
    "max_voltage": dict(type="float", required=False, empty=False, default=3.3),
}


class Sensor(GenericSensor):
    """
    Implementation of MCP3xxx ADC sensor via gpiozero.
    """

    # pylint: disable=too-many-locals,too-many-branches
    def setup_module(self) -> None:
        """
        Init the mcp on SPI CEx
        """

        # read config values
        sensor_spi_port: int = self.config["spi_port"]
        sensor_spi_device: int = self.config["spi_device"]
        sensor_type: str = self.config["type"].upper()
        sensor_channel: int = self.config["channel"]
        sensor_differential: bool = self.config["differential"]
        sensor_max_voltage: float = self.config["max_voltage"]

        # import sensor supertype
        # pylint: disable=import-outside-toplevel,import-error
        from gpiozero import AnalogInputDevice  # type: ignore

        # init the sensor by type
        self.mcp: AnalogInputDevice

        if sensor_type == "MCP3001":
            # pylint: disable=import-outside-toplevel,import-error
            from gpiozero import MCP3001

            self.mcp = MCP3001(
                max_voltage=sensor_max_voltage,
                port=sensor_spi_port,
                device=sensor_spi_device,
            )
        elif sensor_type == "MCP3002":
            # pylint: disable=import-outside-toplevel,import-error
            from gpiozero import MCP3002

            self.mcp = MCP3002(
                channel=sensor_channel,
                differential=sensor_differential,
                max_voltage=sensor_max_voltage,
                port=sensor_spi_port,
                device=sensor_spi_device,
            )
        elif sensor_type == "MCP3004":
            # pylint: disable=import-outside-toplevel,import-error
            from gpiozero import MCP3004

            self.mcp = MCP3004(
                channel=sensor_channel,
                differential=sensor_differential,
                max_voltage=sensor_max_voltage,
                port=sensor_spi_port,
                device=sensor_spi_device,
            )
        elif sensor_type == "MCP3008":
            # pylint: disable=import-outside-toplevel,import-error
            from gpiozero import MCP3008

            self.mcp = MCP3008(
                channel=sensor_channel,
                differential=sensor_differential,
                max_voltage=sensor_max_voltage,
                port=sensor_spi_port,
                device=sensor_spi_device,
            )
        elif sensor_type == "MCP3201":
            # pylint: disable=import-outside-toplevel,import-error
            from gpiozero import MCP3201

            self.mcp = MCP3201(
                max_voltage=sensor_max_voltage,
                port=sensor_spi_port,
                device=sensor_spi_device,
            )
        elif sensor_type == "MCP3202":
            # pylint: disable=import-outside-toplevel,import-error
            from gpiozero import MCP3202

            self.mcp = MCP3202(
                channel=sensor_channel,
                differential=sensor_differential,
                max_voltage=sensor_max_voltage,
                port=sensor_spi_port,
                device=sensor_spi_device,
            )
        elif sensor_type == "MCP3204":
            # pylint: disable=import-outside-toplevel,import-error
            from gpiozero import MCP3204

            self.mcp = MCP3204(
                channel=sensor_channel,
                differential=sensor_differential,
                max_voltage=sensor_max_voltage,
                port=sensor_spi_port,
                device=sensor_spi_device,
            )
        elif sensor_type == "MCP3208":
            # pylint: disable=import-outside-toplevel,import-error
            from gpiozero import MCP3208

            self.mcp = MCP3208(
                channel=sensor_channel,
                differential=sensor_differential,
                max_voltage=sensor_max_voltage,
                port=sensor_spi_port,
                device=sensor_spi_device,
            )
        elif sensor_type == "MCP3301":
            # pylint: disable=import-outside-toplevel,import-error
            from gpiozero import MCP3301

            self.mcp = MCP3301(
                max_voltage=sensor_max_voltage,
                port=sensor_spi_port,
                device=sensor_spi_device,
            )
        elif sensor_type == "MCP3302":
            # pylint: disable=import-outside-toplevel,import-error
            from gpiozero import MCP3302

            self.mcp = MCP3302(
                channel=sensor_channel,
                differential=sensor_differential,
                max_voltage=sensor_max_voltage,
                port=sensor_spi_port,
                device=sensor_spi_device,
            )
        elif sensor_type == "MCP3304":
            # pylint: disable=import-outside-toplevel,import-error
            from gpiozero import MCP3304

            self.mcp = MCP3304(
                channel=sensor_channel,
                differential=sensor_differential,
                max_voltage=sensor_max_voltage,
                port=sensor_spi_port,
                device=sensor_spi_device,
            )
        elif sensor_type == "MCP3308":
            # pylint: disable=import-outside-toplevel,import-error
            from gpiozero import MCP3308

            self.mcp = MCP3308(
                channel=sensor_channel,
                differential=sensor_differential,
                max_voltage=sensor_max_voltage,
                port=sensor_spi_port,
                device=sensor_spi_device,
            )
        else:
            raise RuntimeConfigError("Unsupported MCP type: %s" % sensor_type)

    def get_value(self, sens_conf: ConfigType) -> SensorValueType:
        """
        Get the analog value from the adc for the configured channel
        """
        # Returns an float between 0 and 1 (or -1 to +1 for certain devices operating in
        # differential mode)
        return float(self.mcp.value)



"""
DockerPi 4 Channel Relay GPIO
"""
from typing import Optional
from mqtt_io.modules.gpio import GenericGPIO, PinDirection, PinPUD
from ...types import ConfigType, PinType

REQUIREMENTS = ("smbus",)

CONFIG_SCHEMA = {
    "i2c_bus_num": {
        "type": "integer",
        "required": True,
        "empty": False,
        "default": 1,
    },
    "dev_addr": {
        "type": "integer",
        "required": True,
        "empty": False,
        "default": 0x11,
    },  # 0x10, 0x11, 0x12, 0x13
}

ON = 0x01
OFF = 0x00
DIRECTIONS = None
PULLUPS = None


class GPIO(GenericGPIO):
    """
    Implementation of DockerPi 4 chanel relay.
    """

    def setup_module(self) -> None:
        # pylint: disable=import-outside-toplevel,import-error
        import smbus as gpio  # type: ignore

        addr = self.config["dev_addr"]
        bus = self.config["i2c_bus_num"]
        self.bus = gpio.SMBus(bus)
        self.address = addr

    def setup_pin(
        self,
        pin: PinType,
        direction: PinDirection,
        pullup: PinPUD,
        pin_config: ConfigType,
        initial: Optional[str] = None,
    ) -> None:
        print(
            "setup_pin(pin=%r, direction=%r, pullup=%r, pin_config=%r, initial=%r)"
            % (pin, direction, pullup, pin_config, initial)
        )
        if initial == "high":
            self.set_pin(pin, True)
        elif initial == "low":
            self.set_pin(pin, False)

    def set_pin(self, pin: PinType, value: bool) -> None:
        """Turn on/off relay number self.address"""
        self.bus.write_byte_data(self.address, pin, ON if value else OFF)

    def get_pin(self, pin: PinType) -> bool:
        return bool(self.bus.read_byte_data[self.address, pin])

    def cleanup(self) -> None:
        self.bus.close()



"""
INA219 DC current sensor
"""

# Mandatory:
# - chip_addr
# - shunt_ohms
# Optional, for more precision:
# - amps (Maximum ampere you expect)
# - range (16 or 32) -> Will be the maximum voltage you want to measure
# - gain (40, 80, 160 or 320) -> maximum shunt voltage (milli volt)
# Optional:
# - low_power: send ina219 to sleep between readings

# Output:
# - power (in watt)
# - current (in ampere)
# - bus_voltage (in volt)
# - shunt_voltage (in milli volt)

import logging
from typing import cast

from ...exceptions import RuntimeConfigError
from ...types import CerberusSchemaType, ConfigType, SensorValueType
from . import GenericSensor

_LOG = logging.getLogger(__name__)

REQUIREMENTS = ("pi-ina219",)
CONFIG_SCHEMA: CerberusSchemaType = {
    "chip_addr": dict(type="integer", required=True),
    "shunt_ohms": dict(type="float", required=False, empty=False, default=100),
    "max_amps": dict(type="float", required=False, empty=False),
    "voltage_range": dict(
        type="integer", required=False, empty=False, allowed=[16, 32], default=32
    ),
    "gain": dict(
        type="string",
        required=False,
        empty=False,
        coerce=lambda x: str(x).upper(),
        default="AUTO",
        allowed=["AUTO", "1_40MV", "2_80MV", "4_160MV", "8_320MV"],
    ),
    "low_power": dict(type="boolean", required=False, default=False),
}


class Sensor(GenericSensor):
    """
    Implementation of Sensor class for the INA219 sensor.
    """

    SENSOR_SCHEMA: CerberusSchemaType = {
        "type": dict(
            type="string",
            required=False,
            empty=False,
            default="power",
            allowed=["power", "current", "bus_voltage", "shunt_voltage"],
        )
    }

    def setup_module(self) -> None:
        # pylint: disable=import-outside-toplevel,attribute-defined-outside-init
        # pylint: disable=import-error,no-member
        from ina219 import INA219  # type: ignore

        self.ina = INA219(
            self.config["shunt_ohms"],
            max_expected_amps=self.config.get("max_amps"),
            address=self.config["chip_addr"],
        )

        ## Configure ina sensor with range and gain from config or default
        vrange = {16: self.ina.RANGE_16V, 32: self.ina.RANGE_32V}[
            self.config["voltage_range"]
        ]
        gain = getattr(self.ina, f"GAIN_{self.config['gain']}")

        self.ina.configure(vrange, gain)

    def get_value(self, sens_conf: ConfigType) -> SensorValueType:
        # pylint: disable=import-outside-toplevel,import-error
        from ina219 import DeviceRangeError  # type: ignore

        sens_type = sens_conf["type"]

        if self.config["low_power"]:
            self.ina.wake()

        try:
            if sens_type == "power":
                return cast(float, self.ina.power()) / 1000
            if sens_type == "current":
                return cast(float, self.ina.current()) / 1000
            if sens_type == "bus_voltage":
                return cast(float, self.ina.voltage())
            if sens_type == "shunt_voltage":
                return cast(float, self.ina.shunt_voltage())
        except DeviceRangeError:
            _LOG.exception("Current out of device range with specified shunt resistor")
            return None
        finally:
            if self.config["low_power"]:
                self.ina.sleep()

        raise RuntimeConfigError(
            "ina219 sensor '%s' was not configured to return an accepted value"
            % sens_conf["name"]
        )



"""
GPIO Zero
"""
import logging
from functools import partial
from typing import Optional, Any, Callable, List, Dict, TYPE_CHECKING, cast

from mqtt_io.modules.gpio import (
    GenericGPIO,
    PinDirection,
    PinPUD,
    InterruptSupport,
    InterruptEdge,
)
from mqtt_io.types import PinType, ConfigType

if TYPE_CHECKING:
    # pylint: disable=import-error
    import gpiozero  # type: ignore

_LOG = logging.getLogger(__name__)
REQUIREMENTS = ("gpiozero",)


class GPIO(GenericGPIO):
    """
    Implementation of GPIO class for gpiozero
    """

    INTERRUPT_SUPPORT = InterruptSupport.SOFTWARE_CALLBACK
    PIN_SCHEMA = {
        "kwargs": {
            "type": "dict",
            "required": False,
            "default": {},
        }
    }
    INPUT_SCHEMA = {"class": {"type": "string", "required": False, "default": "Button"}}
    OUTPUT_SCHEMA = {"class": {"type": "string", "required": False, "default": "LED"}}

    def setup_module(self) -> None:
        # pylint: disable=import-outside-toplevel,import-error
        import gpiozero

        self.io = gpiozero
        self._in_pins: Dict[PinType, gpiozero.InputDevice] = {}
        self._out_pins: Dict[PinType, gpiozero.OutputDevice] = {}
        self.pullup_map = {
            PinPUD.OFF: None,
            PinPUD.UP: True,
            PinPUD.DOWN: False,
        }

    def setup_pin(
        self,
        pin: PinType,
        direction: PinDirection,
        pullup: PinPUD,
        pin_config: ConfigType,
        initial: Optional[str] = None,
    ) -> None:

        if direction == PinDirection.OUTPUT:
            pin_config.setdefault("initial_value", initial)
            cls = getattr(self.io, pin_config.get("class", "LED"))
            self._out_pins[pin] = cls(pin, **pin_config.get("kwargs", {}))
        elif direction == PinDirection.INPUT:
            cls = getattr(self.io, pin_config.get("class", "Button"))
            pin_config.setdefault("pull_up", self.pullup_map[pullup])
            pin_config.setdefault(
                "active_state", True if pin_config["pull_up"] is None else None
            )
            pin_config.setdefault("initial_value", initial)
            self._in_pins[pin] = cls(pin, **pin_config.get("kwargs", {}))
        else:
            raise ValueError("Invalid PinDirection")

    def set_pin(self, pin: PinType, value: bool) -> None:
        if value:
            self._out_pins[pin].on()
        else:
            self._out_pins[pin].off()

    def get_pin(self, pin: PinType) -> bool:
        return cast(bool, self._in_pins[pin].is_active)

    def setup_interrupt_callback(
        self,
        pin: PinType,
        edge: InterruptEdge,
        in_conf: ConfigType,
        callback: Callable[[List[Any], Dict[Any, Any]], None],
    ) -> None:
        _LOG.debug(
            "Added interrupt to gpiozero Pi pin '%s' with callback '%s'", pin, callback
        )

        if edge in (InterruptEdge.BOTH, InterruptEdge.RISING):
            self._in_pins[pin].when_activated = partial(callback, True)
        if edge in (InterruptEdge.BOTH, InterruptEdge.FALLING):
            self._in_pins[pin].when_deactivated = partial(callback, False)
        self.interrupt_edges[pin] = edge

    def get_interrupt_value(self, pin: PinType, *args: Any, **kwargs: Any) -> bool:
        return cast(bool, args[0])



"""
PN532 NFC/RFID reader
"""

# `device` value should be set according to nfcpy manual:
#  https://nfcpy.readthedocs.io/en/latest/topics/get-started.html#open-a-local-device
# `tty` - tries all serial ports and drivers and uses first one
# `tty:S0:pn532` - uses pn532 reader on /dev/ttyS0
# `ttyUSB0` - can be used when pn532 is connected via USB UART module
# `usb` - opens first compatible usb reader
#
# List of supported devices: https://nfcpy.readthedocs.io/en/latest/overview.html#supported-devices

from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from mqtt_io.types import ConfigType

from . import GenericStream

if TYPE_CHECKING:
    from nfc.tag import Tag  # type: ignore

REQUIREMENTS = ("nfcpy",)
CONFIG_SCHEMA = {
    "device": {"type": "string", "required": True, "empty": False, "default": "tty"},
}


class Stream(GenericStream):
    """
    Implementation of stream module for reading PM532 tags via UART
    """

    def __init__(self, config: ConfigType):
        self.last_seen_tag: Optional[Tag] = None
        super().__init__(config)

    def setup_module(self) -> None:
        # pylint: disable=import-outside-toplevel,import-error,attribute-defined-outside-init
        from nfc import ContactlessFrontend  # type: ignore

        self.clf = ContactlessFrontend(self.config["device"])

    def read(self) -> Optional[bytes]:
        # pylint: disable=import-outside-toplevel,import-error,attribute-defined-outside-init
        from binascii import hexlify

        self.clf.connect(rdwr={"on-connect": self.__connected})
        if self.last_seen_tag is None:
            return None
        return hexlify(self.last_seen_tag.identifier)

    def write(self, data: bytes) -> None:
        pass

    def cleanup(self) -> None:
        self.clf.close()

    def __connected(self, tag: Tag) -> bool:
        self.last_seen_tag = tag
        return False
